# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import absolute_import

import gc
import ssl
import time
import errno
import socket
import select
import weakref
import logging
import threading
from sjrpc.core import RpcConnection

import pyev


class RpcServer(object):

    """ Base class for all RpcServer classes.
    """

    DEFAULT_TCP_BACKLOG = 30
    NONBLOCKING_ERRORS = (errno.EAGAIN, errno.EWOULDBLOCK)

    def __init__(self, sock, loop=None, conn_args=(), conn_kw={}):
        self._clients = set()
        self.logger = logging.getLogger('sjrpc')
        if loop is None:
            self.loop = pyev.Loop()
        else:
            self.loop = loop
        self._conn_args = conn_args
        self._conn_kw = conn_kw
        self._sock = sock
        self._sock_watcher = self.loop.io(self._sock, pyev.EV_READ, self._handle)
        self._sock_watcher.start()

    @classmethod
    def from_addr(cls, addr, port, tcp_backlog=None, *args, **kwargs):
        sock = socket.socket()
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((addr, port))
        sock.setblocking(False)
        if tcp_backlog is None:
            tcp_backlog = RpcServer.DEFAULT_TCP_BACKLOG
        sock.listen(tcp_backlog)
        return cls(sock, *args, **kwargs)

    def _wrap(self, sock):
        """ Wrap the socket into the :class:`RpcConnection` and return it.
        """
        return RpcConnection(sock, self.loop, *self._conn_args, **self._conn_kw)

    def _handle(self, watcher, revents):
        # Collect offline connections:
        self._collect_connection()
        while True:
            try:
                sock, address = self._sock.accept()
            except socket.error as err:
                if err.errno in self.NONBLOCKING_ERRORS:
                    return
                if err.errno in (errno.ECONNABORTED, errno.EPROTO, errno.EINTR):
                    pass
                else:
                    self.logger.warn('Error while accepting client: %s',
                                     err.strerror)
                    return

            self.logger.info('New incoming connection from %s:%s', *address)
            conn = self._wrap(sock)
            self.register(conn)

    def _collect_connection(self):
        """ Collect the offlines connections.
        """
        for conn in self._clients.copy():
            if not conn:
                self._clients.remove(conn)

#
# Public methods:
#

    def register(self, conn):
        """ Register a new connection on this server.

        :param conn: the connection to register.
        """
        self._clients.add(conn)

    def unregister(self, conn, shutdown=False):
        """ Unregister the specified client from this server.

        If shutdown is specified, client is shutdown before to be unregistered.

        :param conn: the connection to unregister
        :param shutdown: shutdown or not the connection before to register
        """
        if conn in self._clients:
            if shutdown:
                conn.shutdown()

    def run(self):
        """ Run the :class:`RpcServer`.
        """
        self.loop.start()

    def shutdown(self):
        """ Shutdown the :class:`RpcServer` instance.
        """
        self.logger.info('Shutdown requested')
        self._sock_watcher.stop()
        for client in self._clients.copy():
            self.unregister(client, shutdown=True)
        self.loop.stop(pyev.EVBREAK_ALL)


class SSLRpcServer(RpcServer):

    """ SSL version of the RpcServer.
    """

    def __init__(self, sock, certfile, keyfile, loop=None, *args, **kwargs):
        super(SSLRpcServer, self).__init__(sock, loop, *args, **kwargs)
        self._certfile = certfile
        self._keyfile = keyfile

    def _wrap(self, sock):
        sock = ssl.wrap_socket(sock, server_side=True,
                               keyfile=self._keyfile,
                               certfile=self._certfile,
                               ssl_version=ssl.PROTOCOL_TLSv1,
                               do_handshake_on_connect=True)
        return super(SSLRpcServer, self)._wrap(sock)
